@@ -21,3 +21,8 @@ def run_cmd(cmd: list[str]):
         import pty
 
         pty.spawn(cmd, lambda fd: os.read(fd, 1024))
+
+
+def raise_error(txt):
+    typer.echo(typer.style(txt, fg="red"), err=txt)
+    raise typer.Exit(1)
